@@ -3,7 +3,7 @@ package YAHC;
 use strict;
 use warnings;
 
-our $VERSION = '0.032';
+our $VERSION = '0.033';
 
 use EV;
 use Time::HiRes;
@@ -44,6 +44,9 @@ sub YAHC::State::READING                 () { 30  }
 sub YAHC::State::USER_ACTION             () { 35  }
 sub YAHC::State::COMPLETED               () { 100 } # terminal state
 
+sub YAHC::SocketCache::GET               () { 1 }
+sub YAHC::SocketCache::STORE             () { 2 }
+
 use constant {
     # TCP_READ_CHUNK should *NOT* be lower than 16KB because of SSL things.
     # https://metacpan.org/pod/distribution/IO-Socket-SSL/lib/IO/Socket/SSL.pod
@@ -73,6 +76,7 @@ our @EXPORT_OK = qw/
     yahc_conn_attempts_left
     yahc_conn_socket_cache_id
     yahc_conn_register_error
+    yahc_conn_user_data
 /;
 
 our %EXPORT_TAGS = (all => \@EXPORT_OK);
@@ -93,6 +97,7 @@ sub new {
     # and more importantly to share index within the list
     $args->{_target}  = _wrap_host(delete $args->{host})             if $args->{host};
     $args->{_backoff} = _wrap_backoff(delete $args->{backoff_delay}) if $args->{backoff_delay};
+    $args->{_socket_cache} = _wrap_socket_cache(delete $args->{socket_cache}) if $args->{socket_cache};
 
     my %storage;
     my $self = bless {
@@ -101,7 +106,6 @@ sub new {
         storage             => \%storage,
         debug               => delete $args->{debug} || $ENV{YAHC_DEBUG} || 0,
         keep_timeline       => delete $args->{keep_timeline} || $ENV{YAHC_TIMELINE} || 0,
-        socket_cache        => delete $args->{socket_cache},
         pool_args           => $args,
     }, $class;
 
@@ -129,7 +133,7 @@ sub request {
         if exists $self->{connections}{$conn_id};
 
     my $pool_args = $self->{pool_args};
-    do { $request->{$_} ||= $pool_args->{$_} if $pool_args->{$_} } foreach (qw/host port scheme
+    do { $request->{$_} ||= $pool_args->{$_} if $pool_args->{$_} } foreach (qw/host port scheme head
                                                                                request_timeout connect_timeout
                                                                                drain_timeout lifetime_timeout/);
     if ($request->{host}) {
@@ -146,9 +150,16 @@ sub request {
         $request->{_backoff} = $pool_args->{_backoff};
     }
 
+    if ($request->{socket_cache}) {
+        $request->{_socket_cache} = _wrap_socket_cache($request->{socket_cache});
+    } elsif ($pool_args->{_socket_cache}) {
+        $request->{_socket_cache} = $pool_args->{_socket_cache};
+    }
+
     my $scheme = $request->{scheme} ||= 'http';
     my $debug = delete $request->{debug} || $self->{debug};
     my $keep_timeline = delete $request->{keep_timeline} || $self->{keep_timeline};
+    my $user_data = delete $request->{user_data};
 
     my $conn = {
         id          => $conn_id,
@@ -161,6 +172,8 @@ sub request {
         ($debug                   ? (debug => $debug) : ()),
         ($keep_timeline           ? (keep_timeline => $keep_timeline) : ()),
         ($debug || $keep_timeline ? (debug_or_timeline => 1) : ()),
+        (defined $user_data       ? (user_data => $user_data) : ()),
+        pid         => $$,
     };
 
     my %callbacks;
@@ -210,18 +223,12 @@ sub break {
     $self->{loop}->break(EV::BREAK_ONE)
 }
 
-sub socket_cache {
-    my $self = shift;
-    $self->{socket_cache} = $_[0] if @_;
-    return $self->{socket_cache};
-}
-
 ################################################################################
 # Routines to manipulate connections (also user facing)
 ################################################################################
 
 sub yahc_terminal_error {
-    return $_[0] & YAHC::Error::TERMINAL_ERROR() == YAHC::Error::TERMINAL_ERROR() ? 1 : 0;
+    return (($_[0] & YAHC::Error::TERMINAL_ERROR()) == YAHC::Error::TERMINAL_ERROR()) ? 1 : 0;
 }
 
 sub yahc_reinit_conn {
@@ -287,6 +294,12 @@ sub yahc_conn_url {
            . (defined $request->{query_string} ? ("?" . $request->{query_string}) : "");
 }
 
+sub yahc_conn_user_data {
+    my $conn = shift;
+    $conn->{user_data} = $_[0] if @_;
+    return $conn->{user_data};
+}
+
 ################################################################################
 # Internals
 ################################################################################
@@ -300,7 +313,10 @@ sub _run {
         _log_message('YAHC: reinitializing event loop after forking') if $self->{debug};
         $self->{pid} = $$;
         $self->{loop}->loop_fork;
-        $self->{socket_cache} = {} if defined $self->{socket_cache};
+
+        my $active_connections = grep { $$ != $_->{pid} } values %{ $self->{connections} };
+        warn "YAHC has $active_connections active connections after a fork, consider dropping them!"
+            if $active_connections;
     }
 
     if (defined $until_state) {
@@ -375,45 +391,47 @@ sub _set_init_state {
     # don't move attempt increment before boundary check !!!
     # otherwise we can get off-by-one error in yahc_conn_attempts_left
     my $attempt = ++$conn->{attempt};
-    if ($attempt == 1 || !exists $conn->{request}{_backoff}) {
-        if (_init_helper($self, $conn_id) == 1) {
-            _register_in_timeline($conn, "do attempt on next EV iteration, (iteration=%d)", $self->{loop}->iteration)
-                if exists $conn->{debug_or_timeline};
-
-            # from EV docs:
-            # idle watcher call the callback when there are no other pending
-            # watchers of the same or higher priority. The idle watchers are
-            # being called once per event loop iteration - until stopped.
-            #
-            # so, what we do is we start idle watcher with priority 1 which is
-            # higher then 0 used by all IO watchers. As result, the callback
-            # will be called at the end of this iteration. And others if neccessary.
-
-            my $retry_watcher = $watchers->{retry} ||= $self->{loop}->idle_ns(_get_safe_wrapper($self, $conn, sub {
-                shift->stop; # stop this watcher, _set_init_state will start if neccessary
-                _register_in_timeline($conn, "time for new attempt (iteration=%d)", $self->{loop}->iteration)
-                    if exists $conn->{debug_or_timeline};
-                _set_init_state($self, $conn_id)
-            }));
-
-            $retry_watcher->priority(1);
-            $retry_watcher->start;
-        };
-    } else {
+    if ($attempt > 1 && exists $conn->{request}{_backoff}) {
         my $backoff_delay = eval { $conn->{request}{_backoff}->($conn) };
-
         if (my $error = $@) {
             return _set_user_action_state($self, $conn_id, YAHC::Error::CALLBACK_ERROR() | YAHC::Error::TERMINAL_ERROR(),
                 "exception in backoff callback (close connection): $error");
         };
 
-        $self->{loop}->now_update;
-        _register_in_timeline($conn, "setting backoff_timer to %.3fs", $backoff_delay) if exists $conn->{debug_or_timeline};
-        $watchers->{backoff_timer} = $self->{loop}->timer($backoff_delay, 0, _get_safe_wrapper($self, $conn, sub {
-            _register_in_timeline($conn, "backoff timer of %.3fs expired, time for new attempt", $backoff_delay) if exists $conn->{debug_or_timeline};
-            _set_init_state($self, $conn_id) if _init_helper($self, $conn_id) == 1;
-        }));
+        if ($backoff_delay) {
+            $self->{loop}->now_update;
+            _register_in_timeline($conn, "setting backoff_timer to %.3fs", $backoff_delay) if exists $conn->{debug_or_timeline};
+            $watchers->{backoff_timer} = $self->{loop}->timer($backoff_delay, 0, _get_safe_wrapper($self, $conn, sub {
+                _register_in_timeline($conn, "backoff timer of %.3fs expired, time for new attempt", $backoff_delay) if exists $conn->{debug_or_timeline};
+                _set_init_state($self, $conn_id) if _init_helper($self, $conn_id) == 1;
+            }));
+            return;
+        }
     }
+
+    if (_init_helper($self, $conn_id) == 1) {
+        _register_in_timeline($conn, "do attempt on next EV iteration, (iteration=%d)", $self->{loop}->iteration)
+            if exists $conn->{debug_or_timeline};
+
+        # from EV docs:
+        # idle watcher call the callback when there are no other pending
+        # watchers of the same or higher priority. The idle watchers are
+        # being called once per event loop iteration - until stopped.
+        #
+        # so, what we do is we start idle watcher with priority 1 which is
+        # higher then 0 used by all IO watchers. As result, the callback
+        # will be called at the end of this iteration. And others if neccessary.
+
+        my $retry_watcher = $watchers->{retry} ||= $self->{loop}->idle_ns(_get_safe_wrapper($self, $conn, sub {
+            shift->stop; # stop this watcher, _set_init_state will start if neccessary
+            _register_in_timeline($conn, "time for new attempt (iteration=%d)", $self->{loop}->iteration)
+                if exists $conn->{debug_or_timeline};
+            _set_init_state($self, $conn_id)
+        }));
+
+        $retry_watcher->priority(1);
+        $retry_watcher->start;
+    };
 }
 
 sub _init_helper {
@@ -438,18 +456,17 @@ sub _init_helper {
             if exists $conn->{debug_or_timeline};
 
         my $sock;
-        my $socket_cache = $self->{socket_cache};
-        my $socket_cache_id; $socket_cache_id = yahc_conn_socket_cache_id($conn) if defined $socket_cache;
-        if (defined $socket_cache_id && (my $socket_cache_record = $socket_cache->{$socket_cache_id})) {
-            _register_in_timeline($conn, "reuse socket '$socket_cache_id'") if $conn->{debug_or_timeline};
-            $sock = shift @{ $socket_cache_record };
+        if (my $socket_cache = $request->{_socket_cache}) {
+            $sock = $socket_cache->(YAHC::SocketCache::GET(), $conn);
         }
 
         if (defined $sock) {
+            _register_in_timeline($conn, "reuse socket") if $conn->{debug_or_timeline};
             $watchers->{_fh} = $sock;
             $watchers->{io} = $self->{loop}->io($sock, EV::WRITE, sub {});
             _set_write_state($self, $conn_id);
         } else {
+            _register_in_timeline($conn, "build new socket") if $conn->{debug_or_timeline};
             $sock = _build_socket_and_connect($ip, $port);
             _set_connecting_state($self, $conn_id, $sock);
         }
@@ -723,6 +740,8 @@ sub _set_read_state {
                         return;
                     }
                 }
+            } elsif ($decapitated && ($conn->{request}{method} && $conn->{request}{method} eq 'HEAD')) {
+                _set_user_action_state($self, $conn_id); # We are done reading headers for a HEAD request
             } elsif ($decapitated && length($buf) >= $content_length) {
                 $conn->{response}{body} = (length($buf) > $content_length ? substr($buf, 0, $content_length) : $buf);
                 _set_user_action_state($self, $conn_id);
@@ -851,8 +870,7 @@ sub _close_or_cache_socket {
     my $fh = delete $watchers->{_fh} or return;
     delete $watchers->{io}; # implicit stop
 
-    my $socket_cache = $self->{socket_cache};
-    my $socket_cache_id; $socket_cache_id = yahc_conn_socket_cache_id($conn) if defined $socket_cache;
+    my $socket_cache = $conn->{request}{_socket_cache};
 
     # Stolen from Hijk. Thanks guys!!!
     # We always close connections for 1.0 because some servers LIE
@@ -863,22 +881,26 @@ sub _close_or_cache_socket {
     # server anyway, so BEGONE!
 
     if (   $force_close
-        || !defined $socket_cache_id
+        || !defined $socket_cache
         || (($conn->{request}{proto} || '') eq 'HTTP/1.0')
+        || (($conn->{response}{proto} || '') eq 'HTTP/1.0')
         || (($conn->{response}{head}{Connection} || '') eq 'close'))
     {
-        _register_in_timeline($conn, "drop socket %s", $socket_cache_id || '<noyahc_conn_socket_cache_id>') if $conn->{debug_or_timeline};
-        close($fh);
+        _register_in_timeline($conn, "drop socket") if $conn->{debug_or_timeline};
+        close($fh) if ref($fh) eq 'GLOB'; # checking ref to avoid exception
         return;
     }
 
-    push @{ $socket_cache->{$socket_cache_id} ||= [] }, $fh;
-    _register_in_timeline($conn, "save socket '$socket_cache_id' for later use") if $conn->{debug_or_timeline};
+    _register_in_timeline($conn, "storing socket for later use") if $conn->{debug_or_timeline};
+    eval { $socket_cache->(YAHC::SocketCache::STORE(), $conn, $fh); 1; } or do {
+        yahc_conn_register_error($conn, YAHC::Error::CALLBACK_ERROR(), "Exception in socket_cache callback (ignore error): $@");
+    };
 }
 
 sub yahc_conn_socket_cache_id {
     my $conn = shift;
-    my ($host, $ip, $port, $scheme) = @{ $conn->{selected_target} || [] };
+    return unless defined $conn;
+    my ($host, undef, $port, $scheme) = @{ $conn->{selected_target} || [] };
     return unless $host && $port && $scheme;
     # Use $; so we can use the $socket_cache->{$$, $host, $port} idiom to access the cache.
     return join($;, $$, $host, $port, $scheme);
@@ -947,17 +969,19 @@ sub _build_http_message {
     my $conn = shift;
     my $request = $conn->{request};
     my $path_and_qs = ($request->{path} || "/") . (defined $request->{query_string} ? ("?" . $request->{query_string}) : "");
+    my $has_host = 0;
 
     return join(
         $CRLF,
         ($request->{method} || "GET") . " $path_and_qs " . ($request->{protocol} || "HTTP/1.1"),
-        "Host: " . $conn->{selected_target}[0],
         defined($request->{body}) ? ("Content-Length: " . length($request->{body})) : (),
         defined($request->{head}) && @{ $request->{head} } ? (
             map {
+                $has_host ||= lc($request->{head}[2*$_]) eq 'host';
                 $request->{head}[2*$_] . ": " . $request->{head}[2*$_+1]
             } 0..$#{$request->{head}}/2
         ) : (),
+        !$has_host ? ("Host: " . $conn->{selected_target}[0]) : (),
         "",
         defined($request->{body}) ? $request->{body} : ""
     );
@@ -1028,6 +1052,29 @@ sub _wrap_backoff {
     die "YAHC: unsupported backoff format\n";
 }
 
+sub _wrap_socket_cache {
+    my ($value) = @_;
+    my $ref = ref($value);
+
+    return $value if $ref eq 'CODE';
+    return sub {
+        my ($operation, $conn, $sock) = @_;
+        if ($operation == YAHC::SocketCache::GET()) {
+            my $socket_cache_id = yahc_conn_socket_cache_id($conn) or return;
+            return delete $value->{$socket_cache_id};
+        }
+
+        if ($operation == YAHC::SocketCache::STORE()) {
+            my $socket_cache_id = yahc_conn_socket_cache_id($conn) or return;
+            close(delete $value->{$socket_cache_id}) if exists $value->{$socket_cache_id};
+            $value->{$socket_cache_id} = $sock;
+            return;
+        }
+    } if $ref eq 'HASH';
+
+    die "YAHC: unsupported socket_cache format\n";
+}
+
 sub _call_state_callback {
     my ($self, $conn, $cb_name) = @_;
     my $cb = $self->{callbacks}{$conn->{id}}{$cb_name};
@@ -1663,6 +1710,11 @@ Format "host:port". Default port values are omitted.
 
 Same as C<yahc_conn_target> but return full URL
 
+=head2 yahc_conn_user_data
+
+Let user assosiate arbitrary data with a connection. Be aware of not creating
+cyclic reference!
+
 =head2 yahc_conn_errors
 
 Return errors appeared in given connection. Note that the function returns all
